import random
from typing import Dict, List, Optional
import pandas as pd
import re
from dataflow.prompts.text2sql import SQLVariationPrompt
from tqdm import tqdm
from dataflow.utils.registry import OPERATOR_REGISTRY
from dataflow import get_logger
from dataflow.core import OperatorABC
from dataflow.core import LLMServingABC
from dataflow.utils.storage import DataFlowStorage
from dataflow.utils.text2sql.database_manager import DatabaseManager


@OPERATOR_REGISTRY.register()
class SQLVariationGenerator(OperatorABC):
    def __init__(self, llm_serving: LLMServingABC, 
                 database_manager: DatabaseManager,
                 num_variations: int = 10):
        self.llm_serving = llm_serving
        self.logger = get_logger()
        self.database_manager = database_manager
        self.prompt = SQLVariationPrompt()
        self.num_variations = num_variations
        random.seed(42)

    @staticmethod
    def get_desc(lang):
        if lang == "zh":
            return (
                "该算子生成SQL的变种。\n\n"
                "输入参数：\n"
                "- input_sql_key: SQL列名\n"
                "- input_db_id_key: 数据库ID列名\n\n"
            )
        elif lang == "en":
            return (
                "This operator generates variations of SQL.\n\n"
                "Input parameters:\n"
                "- input_sql_key: The name of the SQL column\n"
                "- input_db_id_key: The name of the database ID column\n\n"
            )
        else:
            return "SQL variation generator for Text2SQL tasks."

    def obtain_db_schema(self, db_manager: DatabaseManager, db_id: str) -> tuple:
        return db_manager.get_table_names_and_create_statements(db_id)

    def obtain_insert_statements(self, db_manager: DatabaseManager, db_id: str, table_names: Optional[List[str]] = None) -> Dict[str, List[str]]:
        return db_manager.get_insert_statements(db_id, table_names, limit=2)

    def parse_response(self, response):
        if not response:
            return ""
                
        pattern = r"```sql\s*(.*?)\s*```"
        sql_blocks = re.findall(pattern, response, re.DOTALL)
            
        if sql_blocks:
            last_sql = sql_blocks[-1].strip()
            return last_sql
        else:
            self.logger.warning("No SQL code block found in the response")
            return ""
    
    def check_column(self, dataframe):
        required_columns = [self.input_sql_key, self.input_db_id_key]
        missing_columns = [col for col in required_columns if col not in dataframe.columns]
        if missing_columns:
            raise ValueError(f"Missing required columns: {missing_columns}")

    def run(self, storage: DataFlowStorage,
            input_sql_key: str = "sql",
            input_db_id_key: str = "db_id"
        ):
        self.logger.info(f"1.Running SQLVariationGenerator...")
        self.input_sql_key = input_sql_key
        self.input_db_id_key = input_db_id_key
        
        self.logger.info(f"2.Reading dataframe...")
        dataframe = storage.read("dataframe")
        self.logger.info(f"3.Dataframe read successfully. dataframe: {dataframe}")
        
        self.logger.info(f"4.Checking columns...")
        self.check_column(dataframe)
        
        self.logger.info(f"5.Generating SQL Variations...")
        original_count = len(dataframe)
        
        self.logger.info(f"6.Building prompts...original_count: {original_count}")
        prompts_and_metadata = []
        original_row_indices = []  # Track which original row each prompt comes from
        
        for row_idx, row in tqdm(dataframe.iterrows(), total=len(dataframe), desc="Generating SQL Variations"):
            try:
                self.logger.info(f"7.Obtaining database schema...")
                table_names, create_statements = self.obtain_db_schema(
                    self.database_manager, row[self.input_db_id_key]
                )
                self.logger.info(f"8.Database schema obtained. table_names: {table_names}, create_statements: {create_statements}")
                if not table_names:
                    self.logger.warning(f"Database {row[self.input_db_id_key]} has no tables")
                    continue  
                
                self.logger.info(f"9.Obtaining insert statements...")
                table_name2insert_statements = self.obtain_insert_statements(
                    self.database_manager, row[self.input_db_id_key], table_names
                )
                self.logger.info(f"10.Insert statements obtained. table_name2insert_statements: {table_name2insert_statements}")
                original_sql = row[self.input_sql_key]
                self.logger.info(f"11.Generating SQL Variations...")
                for _ in range(self.num_variations):
                    insert_statements = []
                    for table_name in table_names:
                        insert_statements.extend(table_name2insert_statements.get(table_name, []))
                    self.logger.info(f"12.Insert statements obtained. insert_statements: {insert_statements}")
                    if len(insert_statements) == 0:
                        db_value_prompt = ""
                    else:
                        if len(insert_statements) > 4:
                            insert_statements = random.sample(insert_statements, 4)
                        self.logger.info(f"13.Insert statements obtained. insert_statements: {insert_statements}")
                        db_value_prompt = self.prompt.insert_stmts_template(
                            insert_statements="\n\n".join(insert_statements)
                        )
                        self.logger.info(f"14.DB value prompt obtained. db_value_prompt: {db_value_prompt}")
                    variation_type = random.randint(0, 5)
                    self.logger.info(f"15.Variation type obtained. variation_type: {variation_type}")
                    variation_prompt = self.prompt.variation_type_prompt(variation_type=variation_type)
                    self.logger.info(f"16.Variation prompt obtained. variation_prompt: {variation_prompt}")
                    prompt = self.prompt.sql_variation_prompt(
                        original_sql=original_sql,
                        schema_str="\n\n".join(create_statements),
                        db_value_prompt=db_value_prompt.strip(),
                        variation_prompt=variation_prompt.strip(),
                        db_engine=self.database_manager.db_type,
                    )
                    self.logger.info(f"17.Prompt obtained. prompt: {prompt}")
                    prompts_and_metadata.append((
                        prompt, 
                        row[self.input_db_id_key]
                    ))
                    original_row_indices.append(row_idx)  # Track the original row index
                    self.logger.info(f"18.Prompts and metadata obtained. prompts_and_metadata: {prompts_and_metadata}")
            except Exception as e:
                self.logger.error(f"Error processing database {row[self.input_db_id_key]}: {e}")
                continue  
        
        self.logger.info(f"===19.Generating responses...")
        if prompts_and_metadata:
            try:
                self.logger.info(f"===20.begin to generate responses...")
                prompts = [prompt for prompt, db_id in prompts_and_metadata]
                self.logger.info(f"===21.Prompts obtained. prompts: {prompts}")
                responses = self.llm_serving.generate_from_input(prompts, system_prompt="")
                self.logger.info(f"===22.Responses obtained. responses: {responses}")
                for i, ((prompt, db_id), response) in enumerate(zip(prompts_and_metadata, responses)):
                    sql = self.parse_response(response)
                    if sql:
                        # 获取原始行数据
                        original_row_idx = original_row_indices[i]
                        original_row = dataframe.iloc[original_row_idx]

                        # 创建新行，复制所有原始数据
                        new_row = original_row.copy()

                        # 只更新SQL字段
                        new_row[self.input_sql_key] = sql
                        
                        # 将新行添加到dataframe中
                        dataframe = pd.concat([dataframe, pd.DataFrame([new_row])], ignore_index=True)
                        self.logger.info(f"===23.New row obtained. new_row: {new_row}")
                        self.logger.info(f"===24.Dataframe updated. dataframe: {dataframe}")
            except Exception as e:
                self.logger.error(f"Error generating SQL variations: {e}")
        self.logger.info(f"===25.begin to save dataframe...")
        
        # 处理NaN值，防止存储时出现类型转换错误
        self.logger.info(f"===25.1.Handling NaN values...")
        for col in dataframe.columns:
            try:
                # 检查列是否包含NaN值
                if dataframe[col].isna().any():
                    self.logger.info(f"Processing column {col} with dtype {dataframe[col].dtype}")
                    
                    if dataframe[col].dtype == 'object':
                        # 对象列，将NaN替换为None
                        dataframe[col] = dataframe[col].fillna(value=None)
                    elif str(dataframe[col].dtype).startswith('int'):
                        # 整数列，将NaN替换为0
                        dataframe[col] = dataframe[col].fillna(value=0).astype('int64')
                    elif str(dataframe[col].dtype).startswith('float'):
                        # 浮点列，将NaN替换为0.0
                        dataframe[col] = dataframe[col].fillna(value=0.0)
                    else:
                        # 其他类型，尝试用None填充
                        self.logger.warning(f"Unknown dtype {dataframe[col].dtype} for column {col}, using None")
                        dataframe[col] = dataframe[col].fillna(value=None)
            except Exception as e:
                self.logger.error(f"Error processing column {col}: {e}")
                # 如果出错，尝试最简单的填充方式
                try:
                    dataframe[col] = dataframe[col].fillna(value=None)
                except Exception as e2:
                    self.logger.error(f"Failed to fill NaN in column {col}: {e2}")
                    # 如果还是失败，跳过这个列
                    continue

        self.logger.info(f"===25.2.NaN handling completed.")

        output_file = storage.write(dataframe)
        self.logger.info(f"===26.Output file obtained. output_file: {output_file}")
        self.logger.info(f"Generated {len(dataframe)} records (original: {original_count}, variations: {len(dataframe) - original_count})")
        return []